require 'set'
require 'ostruct'
require 'base64'
require 'shellwords'
require 'nokogiri'
require 'json'
require 'erb'
require 'open-uri'

require 'stacco/base'

class String
  UpcasedInitialisms = Set['ec2', 'iam', 'elb', 'vpc', 'dhcp', 'aws', 'dns']

  def camelize
    return self if (self != self.downcase)
    self.split('_').map{ |part| UpcasedInitialisms.include?(part) ? part.upcase : part.capitalize }.join
  end
end

class Symbol
  def camelize
    self.to_s.camelize.intern
  end
end

class Stacco::Template
  def initialize
    @parameters = []
    @resources = []
  end

  def udscript(roles, vars)
    roles = roles.map{ |role| role.intern }

    unless roles.include? :backend
      vars = vars.dup
      vars.delete 'wallet_data'
    end

    lns = []
    lns.concat(vars.map do |k, v|
      var_val = v.include?("\0") ? Base64.encode64(v) : v
      "export #{k.to_s.upcase}=#{Shellwords.escape(var_val)}\n"
    end)
    lns.concat Stacco::Resources::CloudInit::CommonScript
    roles.each{ |role| lns.concat Stacco::Resources::CloudInit::RoleScripts[role] }
    lns
  end

  class Path < Array
    def elements_for_json
      self
    end
  end

  class ColonPath < Path
    def to_s
      self.elements_for_json.map{ |sym| sym.to_s.camelize }.join('::')
    end
  end

  class ConcatenatedPath < Path
    def to_s
      self.elements_for_json.map{ |sym| sym.to_s.camelize }.join
    end
  end

  class Var < ConcatenatedPath
    def elements_for_json
      [:stacco] + self
    end
  end

  class StackVar < Var
    def elements_for_json
      [:stacco, :stack] + self
    end
  end

  class Type < ColonPath
    def elements_for_json
      [:aws] + self
    end

    def resource(name = nil)
      Resource[*(self + [name].compact)]
    end
  end

  class Resource < ConcatenatedPath
    def elements_for_json
      [:stacco, :resource] + self
    end
  end

  class Mapping < ConcatenatedPath
    def elements_for_json
      [:stacco, :mapping] + self
    end
  end

  AttachmentTypes = {
    Set[Type[:ec2, :vpc], Type[:ec2, :internet_gateway]] => Type[:ec2, :vpc_gateway_attachment],
  }


  def parameter(name, type, opts = {})
    @parameters.push [name, type, opts]
  end

  def resource(type, opts = {})
    name = type.resource(opts.delete(:name))
    @resources.push [name, type, opts]
    name
  end

  def mapping(name, hsh)
    @mappings.push [name, hsh]
  end

  def attach(resource_a, resource_b, opts = {})
  end

  def render
    tpl = {AWSTemplateFormatVersion: "2010-09-09"}
    tpl[:description] = "Stacco stack"

    tpl[:Mappings] = {}

    @mappings.each do |(name, hsh)|
      name = [name] if name.kind_of?(String) or name.kind_of?(Symbol)
      tpl[:Mappings][Mapping[*name]] = hsh
    end

    tpl[:Parameters] = {}

    @parameters.each do |(name, type, args)|
      name = [name] if name.kind_of?(String) or name.kind_of?(Symbol)
      constraints = {}
      args.merge(type: type.camelize).each do |k,v|
        constraints[k.camelize] = v
      end

      tpl[:Parameters][name] = constraints
    end

    tpl[:Resources] = {}

    @resources.each do |(name, type, args)|
      name = [name] if name.kind_of?(String) or name.kind_of?(Symbol)
      properties = {}
      args.each do |k,v|
        properties[k.camelize] = v
      end

      tpl[:Resources][name] = {"Type" => type, "DependsOn" => [], "Properties" => properties}
    end

    tpl
  end

  def export_for_json(o, is_key = false)
    case o
    when Numeric
      o.to_s
    when TrueClass, FalseClass
      o.to_s
    when NilClass
      raise ArgumentError, o
    when Var
      if is_key
        o.to_s
      else
        {"Ref" => o.to_s}
      end
    when Symbol
      ConcatenatedPath[o].to_s
    when Path
      o.to_s
    when Array
      o.map{ |el| export_for_json(el) }
    when Hash
      nh = {}
      o.each{ |k, v| nh[ export_for_json(k, true) ] = export_for_json(v) }
      nh
    when Regexp
      o.source
    else
      o
    end
  end

  def to_json(opts = {})
    self.export_for_json(self.render).to_json
  end
end

if __FILE__ == $0
  tpl = Stacco::Template.new

  if ARGV.delete('-d')
    require 'pp'
    pp JSON.parse(tpl.to_json)
  else
    puts tpl.to_json
  end
end
